"""
The plot module provides support for visualization of diffusion profile data
and diffusion coefficients data using matplotlib.
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splev
from pydiffusion.utils import SF

# Default label font size
label_fontsize = 13
# Default tick font size
tick_fontsize = 11
# Default legend font size
leg_fontsize = 13


def plot_lim(x1, x2, log=False):
    """
    Automatically generate plot range for diffusion plots

    Parameters
    ----------
    x1, x2 : float
        Plot range, can be concentration or diffusivity values.
    log : bool, optional
        Range in log scale or not.

    Returns
    -------
    x1_lim, x2_lim : float
        Plot range output, used for plt.xlim or plt.set_ylim.
    """
    x1, x2 = min(x1, x2), max(x1, x2)
    if not log:
        x1_lim = (x1*100//1)/100
        x2_lim = (x2*100//1+1)/100
    else:
        x1_lim = 10**np.floor(np.log10(x1))
        x2_lim = 10**(np.floor(np.log10(x2))+1)
    return x1_lim, x2_lim


def profileplot(profile, ax=None, err=None, **kwargs):
    """
    Plot diffusion profiles

    Parameters
    ----------
    profile : DiffProfile
        Diffusion profile object
    ax : matplotlib.Axes
        Default axes used if not specified
    kwargs : kwargs
        Passed to 'matplotlib.pyplot.plot', add label to name the profile.

    Returns
    -------
    matplotlib AxesSubplot

    Examples
    --------
    Plot two profiles on one axis:

    >>> ax = profileplot(profile1, c='b', label='profile1')
    >>> profileplot(profile2, ax, c='r', label='profile2')

    """
    dis, X = profile.dis, profile.X
    clw = {'lw': 2, 'label': profile.name}
    args = {**clw, **kwargs}
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    ax.plot(dis, X, **args)

    # Error analysis result plot
    if err is not None:
        profiles = err.profiles['error']
        disp = np.linspace(dis[0], dis[-1], 1e4)
        Xp = np.zeros((2, len(disp)))
        for i in range(2):
            Xp[i] = splev(disp, profiles[i])
        p = ax.plot(disp, Xp[0], '--', lw=2, label='Error')
        ax.plot(disp, Xp[1], '--', c=p[0].get_color(), lw=2)

    ax.set_xlabel('Distance (micron)', fontsize=label_fontsize)
    ax.set_ylabel('Mole fraction', fontsize=label_fontsize)
    ax.set_xlim(dis.min(), dis.max())
    ax.set_ylim(plot_lim(X.min(), X.max()))
    ax.tick_params(labelsize=tick_fontsize)
    leg = ax.legend(numpoints=1, fontsize=leg_fontsize)
    leg.get_frame().set_linewidth(0.0)
    leg.draggable()
    plt.tight_layout()

    return ax


def SFplot(profile, time, Xlim=[], ax=None, **kwargs):
    """
    Plot Sauer-Fraise calculated diffusion coefficients

    Parameters
    ----------
    profile : DiffProfile
        Diffusion profile object, passed to 'pydiffusion.Dmodel.SF'
    time : float
        Passed to 'pydiffusion.Dmodel.SF'
    Xlim : list
        Passed to 'pydiffusion.Dmodel.SF'
    ax : matplotlib.Axes
        Default axes used if not specified
    kwargs : kwargs
        Passed to 'matplotlib.pyplot.semilogy', , add label to name the D curve.

    Returns
    -------
    matplotlib AxesSubplot
    """
    X = profile.X
    sf = SF(profile, time, Xlim)
    clw = {'marker': '.', 'ls': 'none'}
    if 'label' not in kwargs:
        clw['label'] = profile.name+'_%.1fh_SF' % (time/3600)
    args = {**clw, **kwargs}
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    ax.semilogy(X, sf, **args)
    ax.set_xlabel('Mole fraction', fontsize=label_fontsize)
    ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$', fontsize=label_fontsize)
    ax.set_xlim(plot_lim(X.min(), X.max()))
    ax.tick_params(labelsize=tick_fontsize)
    leg = ax.legend(numpoints=1, fontsize=leg_fontsize)
    leg.get_frame().set_linewidth(0.0)
    leg.draggable()
    plt.tight_layout()

    return ax


def DCplot(diffsys, ax=None, err=None, **kwargs):
    """
    Plot diffusion coefficients

    Parameters
    ----------
    diffsys : DiffProfile
        Diffusion system object
    ax : matplotlib.Axes
        Default axes used if not specified
    err : DiffError
        Error analysis result
    kwargs : kwargs
        Passed to 'matplotlib.pyplot.semilogy', add label to name the D curve.

    Returns
    -------
    matplotlib AxesSubplot

    Compare two diffusion coefficients data on one axis:

    >>> ax = DCplot(dsys1, c='b', label='DC1')
    >>> DCplot(dsys2, ax, c='r', label='DC2')

    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    Np, Xr, fD = diffsys.Np, diffsys.Xr, diffsys.Dfunc
    clw = {'lw': 2, 'label': diffsys.name}
    args = {**clw, **kwargs}
    # Diffusion Coefficients plot
    for i in range(Np):
        Xp = np.linspace(Xr[i, 0], Xr[i, 1], 51)
        Dp = np.exp(splev(Xp, fD[i]))
        if i == 0:
            Dmin, Dmax = min(Dp), max(Dp)
            p = ax.semilogy(Xp, Dp, **args)
            clw_nl = {'c': p[0].get_color(), 'label': '_nolegend_'}
            args_nl = {**args, **clw_nl}
        else:
            Dmin, Dmax = min(Dmin, min(Dp)), max(Dmax, max(Dp))
            ax.semilogy(Xp, Dp, **args_nl)
    plt.tight_layout()

    # Error analysis result plot
    if err is not None:
        loc, errors = err.loc, err.errors
        for i in range(Np):
            pid = np.where((loc >= Xr[i, 0]) & (loc <= Xr[i, 1]))[0]
            Dloc = np.exp(splev(loc[pid], fD[i]))
            if i == 0:
                p = ax.semilogy(loc[pid], Dloc * 10**errors[pid, 0], '--', lw=2, label='Error')
            else:
                p = ax.semilogy(loc[pid], Dloc * 10**errors[pid, 0], '--', lw=2)
            ax.semilogy(loc[pid], Dloc * 10**errors[pid, 1], '--', c=p[0].get_color(), lw=2)

    ax.set_xlim(plot_lim(Xr[0, 0], Xr[-1, 1]))
    ax.set_ylim(plot_lim(Dmin, Dmax, True))
    ax.set_xlabel('Mole fraction', fontsize=label_fontsize)
    ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$', fontsize=label_fontsize)
    ax.tick_params(labelsize=tick_fontsize)
    leg = ax.legend(numpoints=1, fontsize=leg_fontsize)
    leg.get_frame().set_linewidth(0.0)
    leg.draggable()
    plt.tight_layout()

    return ax



"""
pyDiffusion combines tools like diffusion simulation, diffusion data smooth,
forward simulation analysis (FSA), etc. to help people analyze diffusion data
efficiently.
"""
__version__ = '0.2.1'


from pydiffusion.core import DiffProfile, DiffSystem
from pydiffusion.simulation import mphSim
from pydiffusion.smooth import datasmooth
from pydiffusion.plot import profileplot, DCplot, SFplot
from pydiffusion.utils import matanocalc, mesh, step
from pydiffusion.Dtools import SF, Hall, Dmodel, FSA
from pydiffusion.io import read_csv, save_csv



"""
The smooth module provides tools for data smoothing of original diffusion
profile data.
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splev, splrep
from pydiffusion.core import DiffProfile
from pydiffusion.io import ask_input, ita_start, ita_finish
from pydiffusion.utils import check_mono


def movingradius(dis, X, r):
    """
    Data smooth using moving average within a radius range
    The first and last data point is unchanged, data at d is averaged by data
    in the range of (d-r, d+r)

    Parameters
    ----------
    dis, X : numpy.array
        Diffusion profile required to be smoothed
    r : float
        Radius (micron) of moving average
    """
    dmin, dmax = dis[0], dis[-1]
    n = np.size(dis)
    f = splrep(dis, X, k=1)
    Xnew = np.zeros(n)
    for i in range(n):
        h = min(abs(dmin-dis[i]), abs(dmax-dis[i]), r)
        Xnew[i] = np.mean(splev(np.linspace(dis[i]-h, dis[i]+h, 100), f))
    return Xnew


def phasesmooth(dis, X, ax, phn=1):
    """
    Data smooth of a single phase, Using movingradius method.

    Parameters
    ----------
    dis, X : numpy.array
        Diffusion profile within a single phase
    ax : matplotlib.axes.Axes
        Plot axes
    phn : int
        Phase #
    """
    mr_msg = 'No Change: Press ENTER (0 input)\n'\
             'Constant: Enter the constant composition (1 input)\n'\
             'Linear: Enter the start and end composition (2 inputs)\n'\
             'Moving Radius: Start & end composition, smooth radius and times (4 inputs)\n'\
             '(Unchanged end composition can be input by \'-\')\n'

    Xsm = X.copy()
    smoo = True
    while smoo:
        ax.cla()
        ax.plot(dis, Xsm, 'bo')
        ax.set_title('Phase #%i' % phn)
        plt.draw()

        # Zoom in or not
        ipt = ask_input('Zoom in? [n]\n')
        if 'y' in ipt or 'Y' in ipt:
            ax.set_title('Select 2 points to zoom in')
            plt.pause(0.1)
            zm = np.array(plt.ginput(2))[:, 0]
        else:
            zm = [dis[0], dis[-1]]
        zmid = np.where((dis >= zm[0]) & (dis <= zm[1]))[0]

        sm = True
        while sm:
            ax.cla()
            ax.plot(dis[zmid], Xsm[zmid], 'bo')
            ax.set_title('Phase #%i' % phn)
            plt.draw()
            Xsmn = np.copy(Xsm[zmid])
            msg = mr_msg+'Current ends: ['+str(Xsmn[0])+' '+str(Xsmn[-1])+']\n'
            while True:
                ipt = ask_input(msg)
                if len(ipt.split()) in (0, 1, 2, 4):
                    break
                else:
                    print('Wrong inputs!')
            # No change
            if len(ipt.split()) == 0:
                Xsmn = np.copy(Xsm[zmid])
            # Constant
            elif len(ipt.split()) == 1:
                Xsmn[:] = float(ipt)
            # Linear
            elif len(ipt.split()) == 2:
                dis_linear = [dis[zmid][0], dis[zmid][-1]]
                X_linear = [Xsmn[0], Xsmn[-1]]
                ipt = ipt.split()
                for i in range(2):
                    if ipt[i] != '-':
                        X_linear[i] = float(ipt[i])
                f_linear = splrep(dis_linear, X_linear, k=1)
                Xsmn = splev(dis[zmid], f_linear)
            # Moving radius
            else:
                ipt = ipt.split()
                X_new = [Xsmn[0], Xsmn[-1]]
                for i in range(2):
                    if ipt[i] != '-':
                        X_new[i] = float(ipt[i])
                Xsmn[0], Xsmn[-1] = X_new[0], X_new[1]
                r, t = float(ipt[2]), int(ipt[3])
                for i in range(t):
                    Xsmn = movingradius(dis[zmid], Xsmn, r)
            ax.cla()
            ax.plot(dis[zmid], Xsm[zmid], 'bo', dis[zmid], Xsmn, 'ro')
            ax.set_title('Phase #%i' % phn)
            plt.draw()
            ipt = ask_input('Redo this smooth? (y/[n])')
            sm = True if 'y' in ipt or 'Y' in ipt else False
            if not sm:
                Xsm[zmid] = Xsmn
        ax.cla()
        ax.plot(dis, X, 'bo', dis, Xsm, 'ro')
        ax.set_title('Phase #%i' % phn)
        plt.draw()
        check_mono(dis, Xsm)
        ipt = ask_input('Further smooth for this phase? (y/[n])')
        smoo = True if 'y' in ipt or 'Y' in ipt else False
    return Xsm


def datasmooth(profile, interface=[], n=2000, name=''):
    """
    Data smooth of diffusion profile. The functions use moving radius method
    on each phase. Please do not close any plot window during the process.

    datasmooth() is the first step of the forward simulation analysis (FSA).

    Parameters
    ----------
    profile: DiffProfile
        Diffusion profile data.
    interface : list of float
        Np-1 locations of interfaces for a Np-phase system
    n : int
        Interpolation number of the smoothed profile

    Returns
    -------
    profile : pydiffusion.diffusion.DiffProfile

    Examples
    --------
    Smooth the experimental profile (exp) with known interfaces [200, 300]:

    >>> ds = datasmooth(exp, [200, 300])

    """
    dis, X = profile.dis, profile.X
    if len(dis) != len(X):
        raise ValueError('Nonequal length of distance and composition data')
    try:
        If = np.array(interface)
    except TypeError:
        print('interface must be array_like')
    if If.ndim != 1:
        raise ValueError('interface must be 1d-array')

    fig = plt.figure()
    ax = fig.add_subplot(111)

    If = [dis[0]-0.5] + list(If) + [dis[-1]+0.5]
    Np = len(If)-1
    Ip = [0]*(Np+1)
    disn, Xn = dis.copy(), X.copy()

    # Same distance values
    for i in range(len(disn)-1):
        for j in range(i+1, len(disn)):
            if disn[i] == disn[j]:
                disn[j] += 1e-5*(j-i)
            elif disn[j] > disn[i]:
                break

    ita_start()

    # Apply phasesmooth to each phase
    for i in range(Np):
        pid = np.where((disn > If[i]) & (disn < If[i+1]))[0]
        try:
            Xn[pid] = phasesmooth(disn[pid], Xn[pid], ax, i+1)
        except (ValueError, TypeError) as error:
            ita_finish()
            raise error
    plt.close()

    ita_finish()

    # Create a sharp interface at interface locations
    # Solubility of each phase will be extended a little
    for i in range(1, Np):
        pid = np.where(disn > If[i])[0][0]
        start = max(pid-5, np.where(disn > If[i-1])[0][0])
        end = min(pid+5, np.where(disn < If[i+1])[0][-1])
        if start+2 < pid:
            fX1 = splrep(disn[start:pid], Xn[start:pid], k=2)
        else:
            fX1 = splrep(disn[start:pid], Xn[start:pid], k=1)
        if pid+1 < end:
            fX2 = splrep(disn[pid:end+1], Xn[pid:end+1], k=2)
        else:
            fX2 = splrep(disn[pid:end+1], Xn[pid:end+1], k=1)
        disn = np.insert(disn, pid, [If[i], If[i]])
        Xn = np.insert(Xn, pid, [splev(If[i], fX1), splev(If[i], fX2)])
        Ip[i] = pid+1
    Ip[-1] = len(Xn)
    disni, Xni = disn.copy(), Xn.copy()

    # Interpolation
    if n > 0:
        ni = [int(n*(If[i]-If[0])//(If[-1]-If[0])) for i in range(Np)]+[n]
        disni, Xni = np.zeros(n), np.zeros(n)
        for i in range(Np):
            fX = splrep(disn[Ip[i]:Ip[i+1]], Xn[Ip[i]:Ip[i+1]], k=1)
            disni[ni[i]:ni[i+1]] = np.linspace(disn[Ip[i]], disn[Ip[i+1]-1], ni[i+1]-ni[i])
            Xni[ni[i]:ni[i+1]] = splev(disni[ni[i]:ni[i+1]], fX)

    print('Smooth finished')

    if name == '':
        name = profile.name+'_smoothed'

    return DiffProfile(disni, Xni, If[1:-1], name=name)



"""
The core module gives definition of main classes in pyDiffusion.
"""

import numpy as np
from scipy.interpolate import splrep, splev


class DiffProfile(object):
    """
    Diffusion Profile

    Parameters
    ----------
    dis : array-like
        Distance data in microns.
    X : array-like
        Concentration data in mole fraction (0~1).
    If : list, optional
        N-1 interfaces locations (in micron) for N phases system.
        Default value is [dis[0]-d1, dis[-1]+d2], d1 d2 are small values.
    name : str, optional
        Name of the current diffusion profile.

    Examples
    --------
    Create a diffusion profile with 10 data points. Interfaces at [2.5, 3.5]:

    >>> dis = np.arange(10)
    >>> X = np.linspace(0, 1, 10)
    >>> If = [2.5, 3.5]
    >>> profile = DiffProfile(dis, X, If, name='test')

    """

    def __init__(self, dis, X, If=[], name='Profile'):
        try:
            self.dis = np.array(dis, dtype=float)
            self.X = np.array(X, dtype=float)
        except TypeError:
            print('Can not convert input into 1d-array')

        try:
            self.name = str(name)
        except TypeError:
            print('name must be able to convert to str type')

        if self.dis.ndim != 1 or self.X.ndim != 1:
            raise ValueError('1d data is required')

        if len(self.dis) != len(self.X):
            raise ValueError('length of dis and X is not equal')

        # dis must be ascending
        if not np.all(self.dis[1:] >= self.dis[:-1]):
            self.X = np.array([x for _, x in sorted(zip(self.dis, self.X))])
            self.dis.sort()

        d1, d2 = 0.5*(self.dis[1]-self.dis[0]), 0.5*(self.dis[-1]-self.dis[-2])

        try:
            If = np.array(If, dtype=float)
        except TypeError:
            print('If must be a list or 1d array')

        if If.ndim != 1:
            raise TypeError('If must be a list or 1d array')

        self.If = np.array([self.dis[0]-d1] + list(If) + [self.dis[-1]+d2])
        self.Ip = np.zeros(len(self.If), dtype=int)
        for i in range(1, len(self.Ip)-1):
            self.Ip[i] = np.where(self.dis > self.If[i])[0][0]
        self.Ip[-1] = len(dis)

    def copy(self, dismax=None, Xmax=None):
        """
        Method to copy DiffProfile
        Distance data or concentration data can be reversed

        Parameters
        ----------
        dismax : if given, output distance data = dismax - dis
        Xmax : if given, output concentration data = Xmax - X

        Examples
        --------
        Reverse a diffusion profile with distance:

        >>> profile_r = profile.copy(dismax=1000)

        For Fe-Ni diffusion couple, with known profile for Ni (profile_Ni),
        calculate profile for Fe:

        >>> profile_Fe = profile_Ni.copy(Xmax=1.0)

        """
        if dismax is not None:
            try:
                dis = dismax-self.dis
                If = dismax-self.If
            except TypeError:
                print('dismax must be a number')
        else:
            dis = self.dis
            If = self.If
        if Xmax is not None:
            try:
                X = Xmax-self.X
            except TypeError:
                print('Xmax must be a number')
        else:
            X = self.X
        return DiffProfile(dis=dis, X=X, If=If[1:-1], name=self.name)


class DiffSystem(object):
    """
    Diffusion System with diffusion coefficients modeling

    Parameters
    ----------
    Xr : List with length of 2 or array-like with shape (,2), optional
        Concentration range for each phase, default = [0,1].
        Save Xr.shape[0] to phase number DiffSystem.Np.
    Dfunc : list of tck tuple, optional
        Np of tck tuple describe the diffusion coefficient function for each phase.
    X, DC : array-like, optional
        Use existing data to model Dfunc.
    Xspl : list of array, optional
        Xspl is the reference locations to create Dfunc, useful for forward
        simulation analysis.
    name : str, optional
        Name of the current diffusion system.


    Examples
    --------
    Create a diffusion system with 3 phases, with known solubilities [0, 0.2], [0.4, 0.7] and [0.8, 1.0]:

    >>> X = np.linspace(0, 1, 10)
    >>> DC = np.linspace(1e-14, 1e-13, 10)
    >>> Xr = [[0, 0.2], [0.4, 0.7], [0.8, 1.0]]
    >>> dsys = DiffSystem(Xr=Xr, X=X, DC=DC, name='Dtest')

    """

    def __init__(self, Xr=[0, 1], Dfunc=None, X=None, DC=None, Xspl=None, name='System'):
        try:
            self.Xr = np.array(Xr, dtype=float)
        except TypeError:
            print('Cannot convert Xr to array')
        if self.Xr.shape == (2,):
            self.Xr = np.array([Xr])
            self.Np = 1
        elif self.Xr.shape[1] == 2:
            self.Np = self.Xr.shape[0]
        else:
            raise ValueError('Xr must has a shape (,2) or (2,)')

        try:
            self.name = str(name)
        except TypeError:
            print('name must be able to convert to str type')

        if Dfunc is not None:
            if len(Dfunc) != self.Np:
                raise ValueError('Length of Dfunc must be equal to phase number Np')
            else:
                self.Dfunc = Dfunc
        elif X is not None and DC is not None:
            try:
                X = np.array(X, dtype=float)
                DC = np.array(DC, dtype=float)
            except TypeError:
                print('Can not convert input into 1d-array')
            if X.ndim != 1 or DC.ndim != 1:
                raise ValueError('1d data is required')
            if len(X) != len(DC):
                raise ValueError('length of X and DC is not equal')
            fD = [0]*self.Np
            for i in range(self.Np):
                pid = np.where((X >= self.Xr[i, 0]) & (X <= self.Xr[i, 1]))[0]
                if len(pid) > 2:
                    fD[i] = splrep(X[pid], np.log(DC[pid]), k=2)
                else:
                    fD[i] = splrep(X[pid], np.log(DC[pid]), k=1)
            self.Dfunc = fD
        if Xspl is not None:
            if len(Xspl) != self.Np:
                raise ValueError('Length of Xspl must be equal to phase number Np')
        self.Xspl = Xspl

    def copy(self, Xmax=None):
        """
        Method to copy DiffSystem
        Concentration can be reversed

        Parameters
        ----------
        Xmax : if given, output concentration = Xmax - X

        Examples
        --------
        For a Fe-Ni diffusion couple, with known diffusion coefficients as function
        of Ni (dsys_Ni), calculate diffusion coefficients as funcition of Fe:

        >>> dsys_Fe = dsys_Ni.copy(Xmax=1.0)

        """
        if Xmax is None:
            return DiffSystem(Xr=self.Xr, Dfunc=self.Dfunc, Xspl=self.Xspl, name=self.name)
        else:
            Xr = Xmax-self.Xr.flatten()[::-1]
            Xr = Xr.reshape((self.Np, 2))
            fD = [0]*self.Np
            for i in range(self.Np):
                k = self.Dfunc[-i-1][2]
                X = np.linspace(Xr[i, 0], Xr[i, 1], 30)
                DC = splev(Xmax-X, self.Dfunc[-i-1])
                fD[i] = splrep(X, DC, k=k)
            Xspl = None if self.Xspl is None else Xmax-self.Xspl[::-1]
            return DiffSystem(Xr=Xr, Dfunc=fD, Xspl=Xspl, name=self.name)


class DiffError(object):
    """
    Error analysis result of diffusion system

    Parameters
    ----------
    diffsys : DiffSystem
        Diffusion system object
    loc : array_like
        Locations performed error analysis
    errors : 2d-array
        Error calculated at loc.
    data : dict
        A dictionary contains all profiles data during error analysis.
        data['exp'] : Experimental collected data.
        data['ref'] : Reference simulated profile.
        data['error'] : Profiles that reaches error_cap during analysis.
    """

    def __init__(self, diffsys, loc, errors, profiles=None):
        self.diffsys = diffsys
        self.loc = np.array(loc)
        self.errors = errors
        if profiles is not None:
            self.profiles = profiles



"""
The io module provides support for reading and writing diffusion profile data
and diffusion coefficients data to csv files.
"""

import numpy as np
import pandas as pd
from scipy.interpolate import splev
from pydiffusion.core import DiffProfile, DiffSystem
import matplotlib.pyplot as plt
import threading

# To solve the problem when matplotlib figure freezes when input used
# https://stackoverflow.com/questions/34938593/matplotlib-freezes-when-input-used-in-spyder
prompt = False
promptText = ""
done = False
waiting = False
response = ""

regular_input = input


def threadfunc():
    global prompt
    global done
    global waiting
    global response

    while not done:
        if prompt:
            prompt = False
            response = regular_input(promptText)
            waiting = True


def ask_input(text):
    global waiting
    global prompt
    global promptText

    promptText = text
    prompt = True

    while not waiting:
        plt.pause(1.0)
    waiting = False

    return response


def ita_start():
    global done
    done = False
    thread = threading.Thread(target=threadfunc)
    thread.start()


def ita_finish():
    global done
    done = True


def save_csv(name=None, profile=None, diffsys=None):
    """
    Save diffusion data as csv file.

    Parameters
    ----------
    name : str
        csv file name, default name is the profile name of diffsys name.
    profile : DiffProfile
        DiffProfile to save.
    diffsys : DiffSystem
        DiffSystem to save. diffsys can be saved by itself or with profile.

    Examples
    --------

    >>> save_csv('data.csv', profile, dsys)

    """
    if profile is None and diffsys is None:
        raise ValueError('No data entered')
    if name is not None and not name.endswith('.csv'):
        name += str(name)+'.csv'
    elif profile is None:
        Xr, fD = diffsys.Xr, diffsys.Dfunc
        X, DC = np.array([]), np.array([])
        for i in range(diffsys.Np):
            Xnew = np.linspace(Xr[i, 0], Xr[i, 1], 30)
            Dnew = np.exp(splev(Xnew, fD[i]))
            X = np.append(X, Xnew)
            DC = np.append(DC, Dnew)
        data = pd.DataFrame({'X': X, 'DC': DC})
        if name is None:
            name = diffsys.name+'.csv'
        data.to_csv(name, index=False)
    elif diffsys is None:
        data = pd.DataFrame({'dis': profile.dis, 'X': profile.X})
        if name is None:
            name = profile.name+'.csv'
        data.to_csv(name, index=False)
    else:
        dis, X, Xr, fD = profile.dis, profile.X, diffsys.Xr, diffsys.Dfunc
        DC = np.zeros(len(dis))
        for i in range(diffsys.Np):
            pid = np.where((X >= Xr[i, 0]) & (X <= Xr[i, 1]))[0]
            DC[pid] = np.exp(splev(X[pid], fD[i]))
        data = pd.DataFrame({'dis': dis, 'X': X, 'DC': DC})
        if name is None:
            name = profile.name+'.csv'
        data.to_csv(name, index=False)


def read_csv(filename, Xlim=None, name=''):
    """
    Read diffusion data from csv file.

    Parameters
    ----------
    filename : str
        csv file path.
    Xlim : list (length of 2), optional
        A list to determine the two ends solubilities.
    name : str, optional
        Name the output DiffProfile and DiffSystem

    Returns
    -------
    profile : DiffProfile
        output DiffProfile object.
    diffsys : DiffSystem
        output DiffSystem object.

    Examples
    --------
    Both profile and diffusivity data:

    >>> profile, dsys = read_csv('data.csv')

    Profile data only:

    >>> profile, _ = read_csv('data.csv')

    """
    data = pd.read_csv(filename)
    if 'X' not in data.columns:
        raise ValueError('No column X in csv file')
    X = np.array(data['X'])

    # Auto rename
    if name == '':
        if '/' in filename:
            r = filename.rfind('/')
        elif '\\' in filename:
            r = filename.rfind('\\')
        else:
            r = -1
        if filename.endswith('.csv'):
            name = filename[r+1:-4]
        else:
            name = filename[r+1:]

    if 'dis' in data.columns:
        dis = np.array(data['dis'])
        If = []
        XIf = []
        for i in range(len(dis)-1):
            if dis[i] == dis[i+1]:
                If += [dis[i]]
                XIf += [X[i], X[i+1]]
        profile = DiffProfile(dis, X, If, name=name)
        if Xlim is None:
            XIf = np.array([X[0]] + XIf + [X[-1]])
        else:
            XIf = np.array([Xlim[0]] + XIf + [Xlim[-1]])
        Xr = XIf.reshape((len(XIf)//2, 2))
    if 'DC' in data.columns:
        DC = np.array(data['DC'])
    else:
        X = DC = None
    if Xr is not None:
        diffsys = DiffSystem(Xr, X=X, DC=DC, name=name)
    return profile, diffsys



"""
The Dtools module provides tools to calculate diffusion coefficients based on
a diffusion profile, like Sauer-Freise method, Hall method and Forward
Simulation Analysis (FSA). This module also provides the construction of
DiffSystem by fitting a smooth diffusion coefficient curve based on
Sauer-Fraise calculation results and the tools to adjust it.
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splrep, splev, UnivariateSpline
from scipy.special import erfinv
from pydiffusion.core import DiffSystem
from pydiffusion.utils import disfunc, matanocalc, error_profile, step, automesh, SF
from pydiffusion.io import ita_start, ita_finish, ask_input
from pydiffusion.plot import profileplot, DCplot, SFplot
from pydiffusion.simulation import mphSim


def SauerFreise(profile, time, Xlim=[]):
    """
    Use Sauer-Freise method to calculate and virtualize diffusion coefficients.

    Parameters
    ----------
    profile : DiffProfile
        Diffusion profile.
    time : float
        Diffusion time in seconds.
    Xlim : list (float), optional
        Indicates the left and right concentration limits for calculation.
        Default value = [profile.X[0], profile.X[-1]].

    Returns
    -------
    DC : numpy.array
        Diffusion coefficients.
    """
    SFplot(profile=profile, time=time, Xlim=Xlim)
    DC = SF(profile=profile, time=time, Xlim=Xlim)
    return DC


def Hall(profile, time, Xlim=[], a=0.25):
    """
    Use Hall method to estimate diffusion coefficients nearby concentration limits.

    Parameters
    ----------
    profile : DiffProfile
        Diffusion profile.
    time : float
        Diffusion time in seconds.
    Xlim : list (float), optional
        Indicates the left and right concentration limits for calculation.
        Default value = [profile.X[0], profile.X[-1]].
    a : float, optional
        Potion of solubility range will be plotted to choose the linear fitting range
        of u vs. lambda. default a =0.25, 0 < a < 1.
        e.g. XL = 0, XR = 1, a = 0.25. [0, 0.25] and [0.75, 1] are the plotting range.

    Returns
    -------
    DC_left, DC_right : numpy.array
        Diffusion coefficients by Hall method at left end and right end.
        Note: Only left part of DC_left and right part of DC_right are valid.
    """
    dis, X = profile.dis, profile.X
    [XL, XR] = [X[0], X[-1]] if Xlim == [] else Xlim

    # Select range for plot
    X1, X2 = XL*(1-a)+XR*a, XL*a+XR*(1-a)
    id1, id2 = (np.abs(X-X1)).argmin(), (np.abs(X-X2)).argmin()

    # Calculate lambda & u
    Y = (X-XL)/(XR-XL)
    matano = matanocalc(profile, Xlim)
    lbd = (dis-matano)/np.sqrt(time)/1e6
    u = erfinv(2*Y-1)

    plt.figure('Hall')
    plt.cla()
    plt.title('LEFT side, select 2 points for linear fitting.')
    plt.plot(lbd[:id1], u[:id1], 'b.')
    plt.xlabel('$\lambda$')
    plt.ylabel('u')
    plt.pause(0.1)
    lbd1 = np.array(plt.ginput(2))[:, 0]

    plt.cla()
    plt.title('RIGHT side, select 2 points for linear fitting.')
    plt.plot(lbd[id2:], u[id2:], 'b.')
    plt.xlabel('$\lambda$')
    plt.ylabel('u')
    plt.pause(0.1)
    lbd2 = np.array(plt.ginput(2))[:, 0]

    sp = np.where((lbd < max(lbd1)) & (lbd > min(lbd1)))[0]
    h1, k1 = np.polyfit(lbd[sp], u[sp], 1)
    sp = np.where((lbd < max(lbd2)) & (lbd > min(lbd2)))[0]
    h2, k2 = np.polyfit(lbd[sp], u[sp], 1)

    DC_left = 1/4/h1**2*(1+2*k1/np.sqrt(np.pi)*np.exp(u**2)*Y)
    DC_right = 1/4/h2**2*(1-2*k2/np.sqrt(np.pi)*np.exp(u**2)*(1-Y))
    DC = SF(profile, time, Xlim)

    plt.cla()
    plt.semilogy(X, DC, 'b.')
    plt.semilogy(X[:id1], DC_left[:id1], 'r--', lw=2)
    plt.semilogy(X[id2:], DC_right[id2:], 'r--', lw=2)
    plt.xlabel('Mole fraction', fontsize=15)
    plt.ylabel('Diffusion Coefficients '+'$(m^2/s)$', fontsize=15)
    plt.xlim(X.min(), X.max())

    return DC_left, DC_right


def Dpcalc(X, DC, Xp):
    """
    Based on Sauer-Fraise calculated results, this function provides good
    estimation of diffusion coefficients at location picked.

    Parameters
    ----------
    X, DC : 1d-array
        Sauer-Fraise calculated results.
    Xp : 1d-array
        Locations to estimate diffusion coefficients.

    Returns
    -------
    Dp : 1d-array
        Estimated diffusion coefficients.
    """
    if len(Xp) == 1:
        fD = splrep(X, np.log(DC), k=1)
        Dp = np.exp(splev(Xp, fD))
    else:
        Dp = np.zeros(len(Xp))
        for i in range(len(Xp)):
            mark = np.zeros(2)
            if i == 0:
                mark[0] = Xp[i]
            else:
                mark[0] = Xp[i-1]
            if i == len(Xp)-1:
                mark[1] = Xp[i]
            else:
                mark[1] = Xp[i+1]
            pid = np.where((X > mark[0]) & (X < mark[1]))[0]
            p = np.polyfit(X[pid], np.log(DC[pid]), 2)
            Dp[i] = np.exp(np.polyval(p, Xp[i]))
    return Dp


def Dfunc_spl(Xp, Dp):
    """
    Return a spline function to model diffusion coefficients.
    The function can be constant(1), linear(2) or quadratic(>2) depending on
    the length of Xp.

    Parameters
    ----------
    Xp : 1d-array
        Composition list.
    Dp : 1d-array
        Corresponding diffusion coefficients at Xp.
    """
    if len(Xp) == 1:
        fDC = splrep([Xp[0], Xp[0]*1.01], [np.log(Dp[0]), np.log(Dp[0])], k=1)
    elif len(Xp) == 2:
        fDC = splrep(Xp, np.log(Dp), k=1)
    else:
        fDC = splrep(Xp, np.log(Dp), k=2)
    return fDC


def Dfunc_uspl(X, DC, Xp, Xr):
    """
    Use UnivariateSpline to model diffusion coefficients.

    Parameters
    ----------
    X, DC : 1d-array
        Diffusion coefficients data.
    Xp : 1d-array with shape (1, 2)
        UnivariateSpline range of X.
    Xr : 1d-array with shape (1, 2)
        Expanded range of UnivariateSpline, usually is the phase range.
    """
    pid = np.where((X >= Xp[0]) & (X <= Xp[-1]))[0]
    fDC = UnivariateSpline(X[pid], np.log(DC[pid]), bbox=[Xr[0], Xr[1]], k=2)
    Xf = np.linspace(Xr[0], Xr[1], 30)
    return splrep(Xf, fDC(Xf), k=2)


def Dadjust(profile_ref, profile_sim, diffsys, ph, pp=True, deltaD=None, r=0.02):
    """
    Adjust diffusion coefficient fitting function by comparing simulated
    profile against reference profile. The purpose is to let simulated
    diffusion profile be similar to reference profile.

    Parameters
    ----------
    profile_ref : DiffProfile
        Reference diffusion profile
    profile_sim : DiffProfile
        Simulated diffusion profile
    diffsys : DiffSystem
        Diffusion system
    ph : int
        Phase # to be adjusted, 0 <= ph <= diffsys.Np-1
    Xp : 1d-array
        Reference composition to adjust their corresponding diffusivities.
        If provided, spline function Dfunc must be determined by [Xp, Dp]
        alone, where Dp = exp(Dfunc(Xp)).
    pp : bool, optional
        Point Mode (True) or Phase Mode (False). Point Mode
        adjusts each Dp at Xp by itself. In Phase Mode, all Dp are
        adjusted by the same rate, i.e. the diffusivity curve shape won't
        change.
    deltaD: float, optional
        Only useful at Phase Mode. deltaD gives the rate to change
        diffusion coefficients DC. DC = DC * 10^deltaD
    r : float, optional
        Only useful at Phase Mode, default = 0.02, 0 < r < 1. r gives the
        range to calculate the concentration gradient around X, [X-r, X+r].

    """
    dref, Xref, Ifref = profile_ref.dis, profile_ref.X, profile_ref.If
    dsim, Xsim, Ifsim = profile_sim.dis, profile_sim.X, profile_sim.If

    if ph >= diffsys.Np:
        raise ValueError('Incorrect phase #, 0 <= ph <= %i' % diffsys.Np-1)
    if pp and 'Xspl' not in dir(diffsys):
        raise ValueError('diffsys must have Xspl properties in per-point mode')

    Dfunc, Xr, Np = diffsys.Dfunc[ph], diffsys.Xr[ph], diffsys.Np
    rate = 1

    # If there is phase consumed, increase adjustment rate
    if len(Ifref) != len(Ifsim):
        print('Phase consumed found, increase adjustment rate')
        rate = 2

    idref = np.where((Xref >= Xr[0]) & (Xref <= Xr[1]))[0]
    idsim = np.where((Xsim >= Xr[0]) & (Xsim <= Xr[1]))[0]

    if 'Xspl' in dir(diffsys):
        Xp = diffsys.Xspl[ph]
    else:
        Xp = np.linspace(Xr[0], Xr[1], 30)
    Dp = np.exp(splev(Xp, Dfunc))

    # If this is consumed phase, increase DC by 2 or 10^deltaD
    if len(idsim) == 0:
        Dp = np.exp(splev(Xp, Dfunc))
        if deltaD is None:
            return Dfunc_spl(Xp, Dp*2)
        else:
            return Dfunc_spl(Xp, Dp*10**deltaD)

    dref, Xref = dref[idref], Xref[idref]
    dsim, Xsim = dsim[idsim], Xsim[idsim]

    # Per phase adjustment
    if not pp:
        if deltaD is not None:
            return Dfunc_spl(Xp, Dp*10**deltaD)

        # Calculate deltaD by phase width
        # When it comes to first or last phase, data closed to end limits are not considered
        fdis_ref = disfunc(dref, Xref)
        fdis_sim = disfunc(dsim, Xsim)
        X1, X2 = Xr[0], Xr[1]
        if ph == 0:
            X1 = Xr[0]*0.9 + Xr[1]*0.1
        if ph == Np-1:
            X2 = Xr[0]*0.1 + Xr[1]*0.9
        ref = splev([X1, X2], fdis_ref)
        sim = splev([X1, X2], fdis_sim)
        wref = ref[1]-ref[0]
        wsim = sim[1]-sim[0]
        Dp *= np.sqrt(wref/wsim)
        return Dfunc_spl(Xp, Dp)

    # Point Mode adjustment
    for i in range(len(Xp)):
        # X1, X2 is the lower, upper bound to collect profile data
        # X1, X2 cannot exceed phase bound Xr
        X1, X2 = max(Xp[i]-r, Xr[0]), min(Xp[i]+r, Xr[1])

        # Calculate the gradient inside [X1, X2] by linear fitting
        fdis_ref = disfunc(dref, Xref)
        fdis_sim = disfunc(dsim, Xsim)
        Xf = np.linspace(X1, X2, 10)
        pref = np.polyfit(splev(Xf, fdis_ref), Xf, 1)[0]
        psim = np.polyfit(splev(Xf, fdis_sim), Xf, 1)[0]

        # Adjust DC by gradient difference
        Dp[i] *= (psim/pref)**rate
    return Dfunc_spl(Xp, Dp)


def Dmodel(profile, time, Xspl=None, Xlim=[], output=True, name=''):
    """
    Given the diffusion profile and diffusion time, modeling the diffusion
    coefficients for each phase. Please do not close any plot window during
    the modeling process.

    Dmodel() is the second step of the forward simulation analysis (FSA).

    Parameters
    ----------
    profile : DiffProfile
        Diffusion profile. Multiple phase profile must be after datasmooth to
        identify phase boundaries.
    time : float
        Diffusion time in seconds.
    Xspl : list, optional
        If Xspl is given, Dmodel will be done automatically.
    Xlim : list, optional
        Left and Right limit of diffusion coefficients. Xlim is also passed to
        SF function to calculate diffusion coefficients initially.
    output : bool, optional
        Plot Dmodel result or not. Can be False only if Xspl is given.
    name : str, optional
        Name of the output DiffSystem

    Examples
    --------
    After datasmooth(), a initial diffusion coefficients will be established before
    FSA():

    >>> ds = datasmooth(exp)
    >>> dsys = Dmodel(ds, time)

    """
    if not isinstance(Xlim, list):
        raise TypeError('Xlim must be a list')
    if len(Xlim) != 2 and Xlim != []:
        raise ValueError('Xlim must be an empty list or a list with length = 2')

    # Initial set-up of Xr (phase boundaries)
    dis, X = profile.dis, profile.X
    Xlim = [X[0], X[-1]] if Xlim == [] else Xlim
    DC = SF(profile, time, Xlim)
    Xr = np.array(Xlim, dtype=float)
    for i in range(len(dis)-1):
        if dis[i] == dis[i+1]:
            Xr = np.insert(Xr, -1, [X[i], X[i+1]])
    Np = len(Xr)//2
    Xr.sort()
    Xr = Xr.reshape(Np, 2)
    fD = [0]*Np

    ita_start()

    # Choose Spline or UnivariateSpline
    if Xspl is None or output:
        ax = plt.figure().gca()
        ax.semilogy(X, DC, 'b.')
        ax.set_title('Sauer-Freise result')
        ax.set_xlabel('Mole fraction')
        ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$')
        plt.tight_layout()

    ipt = ask_input('Use Spline (y) or UnivariateSpline (n) to model diffusion coefficients? [y]\n')
    choice = False if 'N' in ipt or 'n' in ipt else True

    # Xspl provided, no need for manually picking Xspl
    if Xspl is not None:
        if len(Xspl) != Np:
            raise ValueError('Xspl must has a length of phase number')

        for i in range(Np):
            pid = np.where((X >= Xr[i, 0]) & (X <= Xr[i, 1]))[0]

            # Spline
            if choice:
                try:
                    Dp = Dpcalc(X, DC, Xspl[i])
                    fD[i] = Dfunc_spl(Xspl[i], Dp)
                except (ValueError, TypeError) as error:
                    ita_finish()
                    raise error

            # UnivariateSpline
            else:
                try:
                    fD[i] = Dfunc_uspl(X, DC, Xspl[i], Xr[i])
                except (ValueError, TypeError) as error:
                    ita_finish()
                    raise error

        print('DC modeling finished, Xspl info:')
        print(Xspl)

        if output:
            ax.cla()
            ax.set_title('DC Modeling Result')
            ax.semilogy(X, DC, 'b.')
            for i in range(Np):
                Xf = np.linspace(Xr[i, 0], Xr[i, 1], 30)
                ax.semilogy(Xf, np.exp(splev(Xf, fD[i])), 'r-')
            ax.set_xlabel('Mole fraction')
            ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$')
            plt.tight_layout()
            plt.pause(0.1)
            plt.show()

        ita_finish()

        return DiffSystem(Xr, Dfunc=fD, Xspl=Xspl)

    Xspl = [0] * Np if choice else None

    for i in range(Np):
        pid = np.where((X >= Xr[i, 0]) & (X <= Xr[i, 1]))[0]

        # Spline
        if choice:
            while True:
                DC_real = [k for k in DC[pid] if not np.isnan(k) and not np.isinf(k)]
                DCmean = np.mean(DC_real)
                for k in pid:
                    if np.isnan(DC[k]) or np.isinf(DC[k]) or abs(np.log10(DC[k]/DCmean)) > 5:
                        DC[k] = DCmean
                ax.cla()
                ax.semilogy(X[pid], DC[pid], 'b.')
                ax.set_xlabel('Mole fraction')
                ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$')
                ax.set_title('Phase %i' % (i+1))
                plt.draw()
                msg = '# of spline points: 1 (constant), 2 (linear), >2 (spline)\n'
                ipt = ask_input(msg+'input # of spline points\n')
                ax.set_title('Phase %i: Select %i points of Spline' % (i+1, int(ipt)))
                plt.pause(0.1)
                Xp = np.array(plt.ginput(int(ipt)))[:, 0]
                try:
                    Dp = Dpcalc(X, DC, Xp)
                    fD[i] = Dfunc_spl(Xp, Dp)
                except (ValueError, TypeError) as error:
                    ita_finish()
                    raise error
                Xspl[i] = list(Xp)
                Xf = np.linspace(Xr[i, 0], Xr[i, 1], 30)
                ax.semilogy(Xf, np.exp(splev(Xf, fD[i])), 'r-', lw=2)
                plt.draw()
                ipt = ask_input('Continue to next phase? [y]')
                redo = False if 'N' in ipt or 'n' in ipt else True
                if redo:
                    break

        # UnivariateSpline
        else:
            while True:
                ax.cla()
                ax.semilogy(X[pid], DC[pid], 'b.')
                ax.set_xlabel('Mole fraction')
                ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$')
                #ax.set_title('Phase %i' % (i+1))
                ax.set_title('Phase %i: Select 2 boundaries for UnivariateSpline' % (i+1))
                plt.draw()
                plt.pause(0.1)
                Xp = np.array(plt.ginput(2))[:, 0]
                #ipt = ask_input('input 2 boundaries for UnivariateSpline\n')
                #Xp = np.array([float(x) for x in ipt.split()])
                try:
                    fD[i] = Dfunc_uspl(X, DC, Xp, Xr[i])
                except (ValueError, TypeError) as error:
                    ita_finish()
                    raise error
                Xf = np.linspace(Xr[i, 0], Xr[i, 1], 30)
                ax.semilogy(Xf, np.exp(splev(Xf, fD[i])), 'r-', lw=2)
                plt.draw()
                ipt = ask_input('Continue to next phase? [y]')
                redo = False if 'N' in ipt or 'n' in ipt else True
                if redo:
                    break

    ita_finish()

    print('DC modeling finished, Xspl info:')
    print(Xspl)

    ax.cla()
    ax.set_title('DC Modeling Result')
    ax.semilogy(X, DC, 'b.')
    for i in range(Np):
        Xf = np.linspace(Xr[i, 0], Xr[i, 1], 30)
        ax.semilogy(Xf, np.exp(splev(Xf, fD[i])), 'r-')
    ax.set_xlabel('Mole fraction')
    ax.set_ylabel('Diffusion Coefficients '+'$(m^2/s)$')
    plt.tight_layout()
    plt.pause(0.1)
    plt.show()

    if name == '':
        name = profile.name+'_%.1fh_Dmodeled' % (time/3600)

    return DiffSystem(Xr, Dfunc=fD, Xspl=Xspl, name=name)


def FSA(profile_exp, profile_sm, diffsys, time, Xlim=[], n=[400, 500], w=None, f=None, alpha=0.3, name=''):
    """
    Forward Simulation Analysis
    Extract diffusion coefficients based on a diffusion profile.
    Please do not close any plot window during the FSA process.

    This is the final step of FSA.

    Parameters
    ----------
    profile_exp : DiffProfile
        Experimental diffusion profile, used for comparison with simulation
        results.
    profile_sm : DiffProfile
        Diffusion profile after data smooth on experimental profile.
    diffsys : DiffSystem
        Diffusion coefficients
    time : float
        Diffusion time in seconds
    Xlim : list (float), optional
        Passed to 'pydiffusion.Dtools.SF', 'pydiffusion.utils.step'.
        Indicates the left and right concentration limits for calculation.
        Default value = [profile.X[0], profile.X[-1]].
    n : list. optional
        Passed to 'pydiffusion.utils.automesh'.
        Meshing number range, default = [400, 500].
    w : list, optional
        Weights of each phase to calculate error.
        Passed to 'pydiffusion.utils.error_profile'.
    f : function of Meshing
        Keyword argument of automesh()
    alpha : float
        Keyword argument of automesh()
    name : str, optional
        Name of the output DiffProfile and DiffSystem

    Returns
    -------
    profile_sim : DiffProfile
        Simulated diffusion profile after FSA.
    diffsys_sim : DiffSystem
        Calculated diffusion efficients by FSA.

    Examples
    --------
    After datasmooth() and Dmodel(), FSA can be performed to calculate accurate diffusion coefficients:

    >>> ds = datasmooth(exp)
    >>> dsys = Dmodel(ds, time)
    >>> fsa = FSA(exp, ds, dsys, time)

    """
    # Create step profile on meshed grids
    dism = automesh(profile=profile_sm, diffsys=diffsys, n=n, f=f, alpha=alpha)
    matano = matanocalc(profile_sm, Xlim)
    if Xlim == [] and profile_sm.X[-1] < profile_sm.X[0]:
        profile_init = step(dism, matano, diffsys, [diffsys.Xr[-1, 1], diffsys.Xr[0, 0]])
    else:
        profile_init = step(dism, matano, diffsys, Xlim)

    # Determine the stop criteria of forward simulations
    error_sm = error_profile(profile_sm, profile_exp)
    ipt = input('Default error = %.6f\nInput the stop criteria of error: [%.6f]\n'
                % (error_sm, error_sm*2))
    error_stop = error_sm*2 if ipt == '' else float(ipt)

    # If there is no Xspl info in diffsys, use Phase Mode
    # else: ask if use Phase or Point Mode
    if diffsys.Xspl is not None:
        ipt = input('Use Phase Mode? [n]\n(The shape of diffusivity curve does not change)\n')
        pp = False if 'y' in ipt or 'Y' in ipt else True
    else:
        pp = False

    if name == '':
        name = profile_exp.name+'_FSA'

    # Diffusion coefficients used for forward simulations
    diffsys_sim = DiffSystem(diffsys.Xr, diffsys.Dfunc, Xspl=diffsys.Xspl, name=name)

    # Plot FSA status
    fig = plt.figure('FSA', figsize=(16, 6))
    ax1, ax2 = fig.add_subplot(121), fig.add_subplot(122)
    profileplot(profile_exp, ax1, ls='none', marker='o', c='b', fillstyle='none')
    profileplot(profile_sm, ax1, ls='-', c='g', lw=1)
    SFplot(profile_sm, time, Xlim, ax2, ls='none', c='b', marker='.')
    DCplot(diffsys_sim, ax2, ls='-', c='r', lw=2)
    plt.draw()
    plt.tight_layout()
    plt.pause(0.1)

    n_sim = 0
    while True:

        # Simulation
        n_sim += 1
        profile_sim = mphSim(profile_init, diffsys_sim, time, name=name)
        error_sim = error_profile(profile_sim, profile_exp, w)
        print('Simulation %i, error = %f(%f)' % (n_sim, error_sim, error_stop))

        # Plot simulation results
        ax1.cla()
        ax2.cla()
        profileplot(profile_exp, ax1, ls='none', marker='o', c='b', fillstyle='none')
        profileplot(profile_sm, ax1, ls='-', c='g', lw=1)
        profileplot(profile_sim, ax1, ls='-', c='r', lw=2)
        SFplot(profile_sm, time, Xlim, ax2, ls='none', c='b', marker='.')
        DCplot(diffsys_sim, ax2, ls='-', c='r', lw=2)
        plt.draw()
        plt.tight_layout()

        # DC adjust
        Dfunc_adjust = [0] * diffsys_sim.Np

        # If error > stop criteria, continue simulation by auto DC adjustment
        if error_sim > error_stop:
            for ph in range(diffsys_sim.Np):
                try:
                    Dfunc_adjust[ph] = Dadjust(profile_sm, profile_sim, diffsys_sim, ph, pp)
                except (ValueError, TypeError) as error:
                    ita_finish()
                    raise error
            diffsys_sim.Dfunc = Dfunc_adjust

        # If error < stop criteria or simulate too many times
        if error_sim <= error_stop or n_sim > 9:

            ita_start()

            # Ask if exit
            ipt = ask_input('Satisfied with FSA? [n]')
            if 'y' in ipt or 'Y' in ipt:
                ita_finish()
                break

            # If use Point Mode
            if diffsys_sim.Xspl is not None:
                ipt = ask_input('Use Point Mode (y) or Phase Mode (n)? [y]')
                pp = False if 'n' in ipt or 'N' in ipt else True
                if pp:
                    for ph in range(diffsys_sim.Np):
                        try:
                            Dfunc_adjust[ph] = Dadjust(profile_sm, profile_sim, diffsys_sim, ph, pp)
                        except (ValueError, TypeError) as error:
                            ita_finish()
                            raise error
                    diffsys_sim.Dfunc = Dfunc_adjust

                    DCplot(diffsys_sim, ax2, ls='-', c='m', lw=2)
                    plt.draw()
                    plt.pause(0.1)
                    ita_finish()
                    continue

            # Phase Mode, ask if use manual input for each phase
            pp = False
            ipt = input('Phase Mode\nManually input for each phase? [n]')
            manual = True if 'y' in ipt or 'Y' in ipt else False
            for ph in range(diffsys_sim.Np):
                if manual:
                    ipt = input('Input deltaD for phase # %i:\n(DC = DC * 10^deltaD, default deltaD = auto)\n' % (ph+1))
                    deltaD = float(ipt) if ipt != '' else None
                else:
                    deltaD = None
                try:
                    Dfunc_adjust[ph] = Dadjust(profile_sm, profile_sim, diffsys_sim, ph, pp, deltaD)
                except (ValueError, TypeError) as error:
                    ita_finish()
                    raise error

            # Apply the adjustment to diffsys_sim
            diffsys_sim.Dfunc = Dfunc_adjust

            DCplot(diffsys_sim, ax2, ls='-', c='m', lw=2)
            plt.draw()
            plt.pause(0.1)
            ita_finish()

    return profile_sim, diffsys_sim



"""
The simulation module contains diffusion simulation function and
simulation-based error analysis tools.
"""

import numpy as np
from scipy.interpolate import splev, splrep
from pydiffusion.core import DiffProfile, DiffError
from pydiffusion.utils import error_profile, DCbias, profilefunc


def sphSim(profile, diffsys, time, output=True, name=''):
    """
    Single-Phase Diffusion Simulation

    Parameters
    ----------
    profile : DiffProfile
        Initial diffusion profile before simulation.
    diffsys : DiffSystem
        Diffusion coefficients.
    time : float
        time in seconds.
    output : boolean, optional
        Print simulation progress, default = True.

    Returns
    -------
    profile : DiffProfile
        Simulated diffusion profile
    """
    if name == '':
        name = diffsys.name+'_%.1fh' % (time/3600)

    dis, Xs = profile.dis.copy()/1e6, profile.X.copy()
    fD = diffsys.Dfunc
    d = dis[1:]-dis[:-1]
    dj = 0.5*(d[1:]+d[:-1])
    t, m = 0.0, 0
    while t < time:
        Xm = 0.5*(Xs[1:]+Xs[:-1])
        DCs = np.exp(splev(Xm, fD[0]))
        dt = min(d**2/DCs/2)
        dt = time-t if t+dt > time else dt*0.95
        t += dt
        m += 1
        Jf = -DCs*(Xs[1:]-Xs[:-1])/d
        Xs[1:-1] = Xs[1:-1]-dt*(Jf[1:]-Jf[:-1])/dj
        Xs[0] -= Jf[0]/d[0]*dt*2
        Xs[-1] += Jf[-1]/d[-1]*dt*2
        if output and np.mod(m, 3e4) == 0:
            print('%.3f/%.3f hrs simulated' % (t/3600, time/3600))
    if output:
        print('Simulation Complete')
    return DiffProfile(dis*1e6, Xs, name=name)


def mphSim(profile, diffsys, time, liquid=0, output=True, name=''):
    """
    Single/Multiple-Phase Diffusion Simulation. Liquid phase can be attached
    at left or right end. For thin film simulation, please set up the
    interface nearby the end in the initial profile.

    Parameters
    ----------
    profile : DiffProfile
        Initial profile before simulation.
    diffsys : DiffSystem
        Diffusion coefficients.
    time : float
        time in seconds.
    liquid : 1, 0 or -1, optional
        Liquid phase provide infinite mass flux during simulation, can only be
        attached at left or right end.
        0 : No liquid phase attached.
        -1 : Liquid phase attached at left.
        1 : Liquid phase attached at right.
    output : boolean, optional
        Print simulation progress, default = True.

    Returns
    -------
    profile : DiffProfile
        Simulated diffusion profile

    Examples
    --------
    With known diffusion coefficients (dsys), simulate profile of 100 hours of diffusion
    for a diffusion couple experiment (initial profile is a step) on a 600 micron grids:

    >>> dis = mesh(0, 600)
    >>> init_profile = step(dis, 300, dsys)
    >>> final_profile = mphSim(init_profile, dsys, 3600*100)

    If liquid phase is attached to the left:

    >>> final_profile = mphSim(init_profile, dsys, 3600*100, liquid=-1)

    """
    dis, Xs = profile.dis.copy()/1e6, profile.X.copy()
    Ip, If = profile.Ip.copy(), profile.If.copy()/1e6
    Np, Xr = diffsys.Np, diffsys.Xr.copy()
    fD = [f for f in diffsys.Dfunc]

    if name == '':
        name = diffsys.name+'_%.1fh' % (time/3600)

    if len(If) != Np+1:
        raise ValueError('Number of phases mismatches between profile and diffusion system')
    if liquid not in [-1, 0, 1]:
        raise ValueError('liquid can only be 0 1 or -1')
    try:
        time = float(time)
    except TypeError:
        print('Wrong type for time variable')

    d = dis[1:]-dis[:-1]
    dj = 0.5*(d[1:]+d[:-1])
    Jf, DCs = np.zeros(len(dis)-1), np.zeros(len(dis)-1)
    JIf = np.zeros([Np+1, 2])
    vIf = np.zeros(Np+1)
    t, m = 0.0, 0
    dt0 = time/1e3

    while t < time:

        Xm = (Xs[:-1]+Xs[1:])/2
        dt = dt0

        # Jf JIf calculation
        # dt limited by simulation stability inside each phases
        for i in range(Np):
            if Ip[i+1] > Ip[i]+1:
                Ipr = np.arange(Ip[i], Ip[i+1]-1)
                DCs[Ipr] = np.exp(splev(Xm[Ipr], fD[i]))
                Jf[Ipr] = -DCs[Ipr]*(Xs[Ipr+1]-Xs[Ipr])/d[Ipr]
                dt = min(dt, min(abs(d[Ipr]**2/DCs[Ipr]/2)))
            if Ip[i+1] == Ip[i]:
                X0 = np.mean(Xr[i])
                JIf[i, 1] = JIf[i+1, 0] = -(Xr[i, 1]-Xr[i, 0])/(If[i+1]-If[i])*np.exp(splev(X0, fD[i]))
            else:
                if i < Np-1:
                    X0 = 0.5*(Xr[i, 1]+Xs[Ip[i+1]-1])
                    JIf[i+1, 0] = -(Xr[i, 1]-Xs[Ip[i+1]-1])/(If[i+1]-dis[Ip[i+1]-1])*np.exp(splev(X0, fD[i]))
                if i > 0:
                    X0 = 0.5*(Xr[i, 0]+Xs[Ip[i]])
                    JIf[i, 1] = -(Xs[Ip[i]]-Xr[i, 0])/(dis[Ip[i]]-If[i])*np.exp(splev(X0, fD[i]))

        # vIf calculation, dt limited by 'No interface passing by'
        for i in range(1, Np):
            vIf[i] = (JIf[i, 1]-JIf[i, 0])/(Xr[i, 0]-Xr[i-1, 1])
            vid = [Ip[i]-2] if Ip[i] > 1 else []
            vid += [Ip[i]] if Ip[i] < len(dis) else []
            dt = min(dt, abs(min(d[vid]/vIf[i])))
            if i > 1 and vIf[i-1] > vIf[i]:
                dt = min(dt, (If[i]-If[i-1])/(vIf[i-1]-vIf[i])/2)

        # dt limited by grid nearby interfaces cannot exceed solubility limit
        for i in range(Np):
            if Ip[i+1] == Ip[i]:
                continue
            elif Ip[i+1] == Ip[i]+1:
                if JIf[i, 1] > JIf[i+1, 0]:
                    dt = min(dt, (Xr[i, 1]-Xs[Ip[i]])*dj[Ip[i]-1]/(JIf[i, 1]-JIf[i+1, 0]))
                elif JIf[i, 1] < JIf[i+1, 0]:
                    dt = min(dt, (Xs[Ip[i]]-Xr[i, 0])*dj[Ip[i]-1]/(JIf[i+1, 0]-JIf[i, 1]))
            else:
                if i < Np-1 and JIf[i+1, 0] < Jf[Ip[i+1]-2]:
                    dt = min(dt, -(Xr[i, 1]-Xs[Ip[i+1]-1])/(JIf[i+1, 0]-Jf[Ip[i+1]-2])*dj[Ip[i+1]-2])
                if i > 0 and Jf[Ip[i]] > JIf[i, 1]:
                    dt = min(dt, (Xs[Ip[i]]-Xr[i, 0])/(Jf[Ip[i]]-JIf[i, 1])*dj[Ip[i]-1])

        dt = time-t if t+dt > time else dt*0.95

        # If first or last phase will be consumed
        if If[1] < dis[1] and vIf[1] < 0:
            dt = min(dt, (dis[0]-If[1])/vIf[1])
        elif If[-2] > dis[-2] and vIf[-2] > 0:
            dt = min(dt, (dis[-1]-If[-2])/vIf[-2])

        t += dt
        m += 1

        # Ficks 2nd law inside each phase
        for i in range(Np):
            if Ip[i+1] == Ip[i]:
                continue
            elif Ip[i+1] == Ip[i]+1:
                Xs[Ip[i]] -= dt*(JIf[i+1, 0]-JIf[i, 1])/dj[Ip[i]-1]
            else:
                if i > 0:
                    Xs[Ip[i]] -= dt*(Jf[Ip[i]]-JIf[i, 1])/dj[Ip[i]-1]
                if i < Np-1:
                    Xs[Ip[i+1]-1] -= dt*(JIf[i+1, 0]-Jf[Ip[i+1]-2])/dj[Ip[i+1]-2]
                if Ip[i+1] > Ip[i]+2:
                    Ipr = np.arange(Ip[i]+1, Ip[i+1]-1)
                    Xs[Ipr] -= dt*(Jf[Ipr]-Jf[Ipr-1])/dj[Ipr-1]

        # Composition changes at first & last grid
        # If there is liquid phase attached, composition unchanged.
        if liquid != -1:
            Xs[0] -= Jf[0]/d[0]*dt
        if liquid != 1:
            Xs[-1] += Jf[-1]/d[-1]*dt

        # If one phase consumed, delete this phase
        if If[1]+vIf[1]*dt <= dis[0]:
            Xs[0] = Xr[1, 0]
            Np -= 1
            Xr, If, Ip, fD = Xr[1:], If[1:], Ip[1:], fD[1:]
            Ip[0] = 0
            JIf = np.zeros([Np+1, 2])
            vIf = np.zeros(Np+1)
            if output:
                print('First phase consumed, %i phase(s) left, time = %.3f hrs' % (Np, t/3600))
        elif If[-2]+vIf[-2]*dt >= dis[-1]:
            Xs[-1] = Xr[-2, 1]
            Np -= 1
            Xr, If, Ip, fD = Xr[:-1], If[:-1], Ip[:-1], fD[:-1]
            Ip[-1] = len(dis)
            JIf = np.zeros([Np+1, 2])
            vIf = np.zeros(Np+1)
            if output:
                print('Last phase consumed, %i phase(s) left, time = %.3f hrs' % (Np, t/3600))

        # Interface move across one grid, passed grid has value change
        for i in range(1, Np):
            If[i] += vIf[i]*dt
            if If[i] < dis[Ip[i]-1]:
                Ip[i] -= 1
                if If[i+1] < dis[Ip[i]+1]:
                    Xs[Ip[i]] = splev(dis[Ip[i]],
                                      splrep([If[i], If[i+1]], Xr[i], k=1))
                else:
                    Xs[Ip[i]] = splev(dis[Ip[i]],
                                      splrep([If[i], dis[Ip[i]+1]],
                                             [Xr[i, 0], Xs[Ip[i]+1]], k=1)
                                      )
            elif If[i] > dis[Ip[i]]:
                Ip[i] += 1
                if If[i-1] > dis[Ip[i]-2]:
                    Xs[Ip[i]-1] = splev(dis[Ip[i]-1],
                                        splrep([If[i-1], If[i]], Xr[i-1], k=1))
                else:
                    Xs[Ip[i]-1] = splev(dis[Ip[i]-1],
                                        splrep([dis[Ip[i]-2], If[i]],
                                               [Xs[Ip[i]-2], Xr[i-1, 1]], k=1)
                                        )

        if output and np.mod(m, 3e4) == 0:
            print('%.3f/%.3f hrs simulated' % (t/3600, time/3600))
    if output:
        print('Simulation Complete')

    # Insert interface informations
    for i in list(range(Np-1, 0, -1)):
        dis = np.insert(dis, Ip[i], [If[i], If[i]])
        Xs = np.insert(Xs, Ip[i], [Xr[i-1, 1], Xr[i, 0]])

    return DiffProfile(dis*1e6, Xs, If[1:-1]*1e6, name=name)


def ErrorAnalysis(profile_exp, profile_init, diffsys, time, loc=10, w=None,
                  r=0.3, efunc=None, accuracy=1e-3, output=False):
    """
    Error analysis of diffusion coefficients through comparison with experimental
    data.

    Parameters
    ----------
    profile_exp : DiffProfile
        Experiment measured diffusion profile, which is used to compared against
        each simulation result.
    profile_init : DiffProfile
        The initial profile for diffusion simulation, usually a step profile.
    diffsys : DiffSystem
        Reference diffusion coefficients. Simulation based on this datasets will
        be the reference for the error analysis.
        Bias will then be applied to this diffusivity datasets before each
        simulation.
    time : float
        Diffusion time in seconds.
    loc : list or int
        loc indicates the locations to perform error analysis. If loc is an
        integer, loc points are selected inside each phase. Each point has
        both positive and negative error to be calculated.
    w : list, optional
        Weights of each phase to calculate error.
        Passed to 'pydiffusion.utils.error_profile'.
    r : float, optional
        The concentration range of bias.
        Passed to 'pydiffusion.utils.DCbias'
    efunc : function, optional
        Function to create bias.
        Passed to 'pydiffusion.utils.DCbias'
    accuracy : float
        Stop criterion of each simulation: within error_cap * (1+-accuracy).
        Low accuracy value may increase simulation times for each point.
    output : boolean, optional
        Print each simulation results, default = False.

    Returns
    -------
    differror : DiffError
        Diffusion error object
    """
    if len(profile_init.If) != diffsys.Np+1:
        raise ValueError('Number of phases mismatches between profile and diffusion system')
    try:
        time = float(time)
    except TypeError:
        print('Wrong type for time variable')

    profile_ref = mphSim(profile_init, diffsys, time, output=output)
    error_ref = error_profile(profile_ref, profile_exp, w)
    ipt = input('Reference error= % .6f. Input cap error: [% .6f]' % (error_ref, error_ref*1.01))
    error_cap = error_ref*1.01 if ipt == '' else float(ipt)
    print('Cap error = % .6f' % error_cap)

    if isinstance(loc, int):
        n = loc
        loc = np.array([])
        for i in range(diffsys.Np):
            loc = np.append(loc, np.linspace(diffsys.Xr[i, 0], diffsys.Xr[i, 1], n))

    dis_compare = np.linspace(profile_init.dis[0], profile_init.dis[-1], 1e4)
    profile_compare = np.zeros((3, len(dis_compare)))
    profile_compare[0] = splev(dis_compare, profilefunc(profile_ref))
    profile_compare[2] = splev(dis_compare, profilefunc(profile_ref))

    errors = np.zeros((len(loc), 2))
    deltaD_init = [0.5, -0.5]
    for i in range(len(loc)):
        X = loc[i]
        for p in range(2):
            n_sim = 0
            deltaD = deltaD_init[p]*1.1
            De, Xe = [0], [error_ref]
            while True:
                n_sim += 1
                diffsys_error = DCbias(diffsys, X, deltaD, r, efunc)
                profile_error = mphSim(profile_init, diffsys_error, time, output=output)
                error_sim = error_profile(profile_error, profile_exp, w)
                if output:
                    print('At %.3f, simulation #%i, deltaD = %f, profile difference = %f(%f)'
                          % (X, n_sim, deltaD, error_sim, error_cap))

                if error_sim <= error_cap:
                    profile_new = profilefunc(profile_error)
                    profile_compare[1] = splev(dis_compare, profile_new)
                    profile_compare[0] = profile_compare[:2].min(0)
                    profile_compare[2] = profile_compare[1:].max(0)

                if error_sim > error_cap*(1-accuracy) and error_sim <= error_cap:
                    break

                if len(De) == 1:
                    if error_sim > error_cap:
                        De += [deltaD]
                        Xe += [error_sim]
                        fe = splrep(Xe, De, k=1)
                        deltaD = float(splev(error_cap, fe))
                    else:
                        De, Xe = [deltaD], [error_sim]
                        deltaD *= 2
                else:
                    j = 1 if error_sim > error_cap else 0
                    De[j], Xe[j] = deltaD, error_sim

                    if (Xe[1]-Xe[0])/abs(De[1]-De[0]) > 10:
                        deltaD, error_sim = De[0], Xe[0]
                        if output:
                            print('Jump between %f and %f' % (De[0], De[1]))
                        break
                    elif n_sim > 6:
                        deltaD = np.mean(De)
                    else:
                        fe = splrep(Xe, De, k=1)
                        deltaD = float(splev(error_cap, fe))
            direction = 'positive' if p == 0 else 'negative'
            print('Error (%s) at %.3f = %f, %i simulations performed, profile difference = %f'
                  % (direction, X, deltaD, n_sim, error_sim))
            errors[i, p] = deltaD
            deltaD_init[p] = deltaD
        profiles = (splrep(dis_compare, profile_compare[0], k=1),
                    splrep(dis_compare, profile_compare[2], k=1))
        data = {}
        data['exp'] = profile_exp
        data['ref'] = profile_ref
        data['error'] = profiles

    print('Error analysis complete')
    differror = DiffError(diffsys, loc, errors, data)

    return differror
